from openfisca_us.model_api import *


class additional_standard_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Additional standard deduction"
    unit = USD
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/63#f"

    def formula(tax_unit, period, parameters):
        std = parameters(period).gov.irs.deductions.standard
        filing_status = tax_unit("filing_status", period)
        filing_statuses = filing_status.possible_values
        blind_head = tax_unit("blind_head", period) * 1
        blind_spouse = tax_unit("blind_spouse", period) * 1
        aged_head = (
            tax_unit("age_head", period) >= std.aged_or_blind.age_threshold
        ) * 1
        aged_spouse = (
            tax_unit("age_spouse", period) >= std.aged_or_blind.age_threshold
        ) * 1
        count_extra_stded = blind_head + blind_spouse + aged_head + aged_spouse
        return count_extra_stded * std.aged_or_blind.amount[filing_status]



from openfisca_us.model_api import *


class self_employment_tax(Variable):
    value_type = float
    entity = Person
    label = "Self-employment tax"
    definition_period = YEAR
    unit = USD

    formula = sum_of_variables(
        ["self_employment_social_security_tax", "self_employment_medicare_tax"]
    )



from openfisca_us.model_api import *


class ordinary_dividend_income(Variable):
    value_type = float
    entity = Person
    label = "Ordinary dividend income"
    unit = USD
    definition_period = YEAR



from openfisca_us.model_api import *


class ctc_refundable_individual_maximum(Variable):
    value_type = float
    entity = Person
    label = "Maximum refundable CTC"
    unit = USD
    documentation = "The maximum refundable CTC for this person."
    definition_period = YEAR
    reference = (
        "https://www.law.cornell.edu/uscode/text/26/24#a",
        "https://www.law.cornell.edu/uscode/text/26/24#h",
        "https://www.law.cornell.edu/uscode/text/26/24#i",
    )

    def formula(person, period, parameters):
        return person("ctc_individual_maximum", period)

    def formula_2018(person, period, parameters):
        amount = person("ctc_individual_maximum", period)
        ctc = parameters(period).gov.irs.credits.ctc
        return min_(amount, ctc.refundable.individual_max)

    def formula_2021(person, period, parameters):
        return person("ctc_child_individual_maximum", period)

    formula_2022 = formula_2018

    formula_2026 = formula



from openfisca_us.model_api import *


class dwks10(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "DWKS10"
    documentation = (
        "search taxcalc/calcfunctions.py for how calculated and used"
    )
    unit = USD

    def formula(tax_unit, period, parameters):
        dwks10_if_gains = add(tax_unit, period, ["dwks6", "dwks9"])
        dwks10_if_no_gains = max_(
            0,
            min_(
                add(tax_unit, period, ["long_term_capital_gains"]),
                tax_unit("c23650", period),
            ),
        ) + add(tax_unit, period, ["non_sch_d_capital_gains"])
        return where(
            tax_unit("hasqdivltcg", period),
            dwks10_if_gains,
            dwks10_if_no_gains,
        )



from io import BytesIO
from zipfile import ZipFile
from openfisca_tools.data import PublicDataset
import pandas as pd
import requests
from tqdm import tqdm
from openfisca_us.data.storage import OPENFISCA_US_MICRODATA_FOLDER


class RawCPS(PublicDataset):
    name = "raw_cps"
    label = "Raw CPS"
    folder_path = OPENFISCA_US_MICRODATA_FOLDER
    is_openfisca_compatible = False

    def generate(self, year: int) -> pd.DataFrame:
        """Generates the raw CPS dataset.

        Args:
            year (int): The year of the raw CPS to use.
        """
        # Files are named for a year after the year the survey represents.
        # For example, the 2020 CPS was administered in March 2021, so it's
        # named 2021.
        file_year = int(year) + 1
        file_year_code = str(file_year)[-2:]

        CPS_URL_BY_YEAR = {
            2020: "https://www2.census.gov/programs-surveys/cps/datasets/2021/march/asecpub21csv.zip"
        }

        if year not in CPS_URL_BY_YEAR:
            raise ValueError(f"No raw CPS data URL known for year {year}.")

        url = CPS_URL_BY_YEAR[year]

        response = requests.get(url, stream=True)
        total_size_in_bytes = int(
            response.headers.get("content-length", 200e6)
        )
        progress_bar = tqdm(
            total=total_size_in_bytes,
            unit="iB",
            unit_scale=True,
            desc="Downloading ASEC",
        )
        if response.status_code == 404:
            raise FileNotFoundError(
                "Received a 404 response when fetching the data."
            )
        try:
            with BytesIO() as file, pd.HDFStore(
                self.file(year), mode="w"
            ) as storage:
                content_length_actual = 0
                for data in response.iter_content(int(1e6)):
                    progress_bar.update(len(data))
                    content_length_actual += len(data)
                    file.write(data)
                progress_bar.set_description("Downloaded ASEC")
                progress_bar.total = content_length_actual
                progress_bar.close()
                zipfile = ZipFile(file)
                with zipfile.open(f"pppub{file_year_code}.csv") as f:
                    storage["person"] = person = pd.read_csv(f).fillna(0)
                with zipfile.open(f"ffpub{file_year_code}.csv") as f:
                    person_family_id = person.PH_SEQ * 10 + person.PF_SEQ
                    family = pd.read_csv(f).fillna(0)
                    family_id = family.FH_SEQ * 10 + family.FFPOS
                    family = family[family_id.isin(person_family_id)]
                    storage["family"] = family
                with zipfile.open(f"hhpub{file_year_code}.csv") as f:
                    person_household_id = person.PH_SEQ
                    household = pd.read_csv(f).fillna(0)
                    household_id = household.H_SEQ
                    household = household[
                        household_id.isin(person_household_id)
                    ]
                    storage["household"] = household
                storage["tax_unit"] = RawCPS._create_tax_unit_table(person)
                storage["spm_unit"] = RawCPS._create_spm_unit_table(person)
        except Exception as e:
            self.remove(year)
            raise ValueError(
                f"Attempted to extract and save the CSV files, but encountered an error: {e} (removed the intermediate dataset)."
            )

    @staticmethod
    def _create_tax_unit_table(person: pd.DataFrame) -> pd.DataFrame:
        TAX_UNIT_COLUMNS = [
            "ACTC_CRD",
            "AGI",
            "CTC_CRD",
            "EIT_CRED",
            "FED_RET",
            "FEDTAX_AC",
            "FEDTAX_BC",
            "MARG_TAX",
            "STATETAX_A",
            "STATETAX_B",
            "TAX_INC",
        ]
        tax_unit_df = person[TAX_UNIT_COLUMNS].groupby(person.TAX_ID).sum()
        tax_unit_df["TAX_ID"] = tax_unit_df.index
        return tax_unit_df

    @staticmethod
    def _create_spm_unit_table(person: pd.DataFrame) -> pd.DataFrame:
        SPM_UNIT_COLUMNS = [
            "ACTC",
            "CAPHOUSESUB",
            "CAPWKCCXPNS",
            "CHILDCAREXPNS",
            "CHILDSUPPD",
            "EITC",
            "ENGVAL",
            "EQUIVSCALE",
            "FAMTYPE",
            "FEDTAX",
            "FEDTAXBC",
            "FICA",
            "GEOADJ",
            "HAGE",
            "HHISP",
            "HMARITALSTATUS",
            "HRACE",
            "MEDXPNS",
            "NUMADULTS",
            "NUMKIDS",
            "NUMPER",
            "POOR",
            "POVTHRESHOLD",
            "RESOURCES",
            "SCHLUNCH",
            "SNAPSUB",
            "STTAX",
            "TENMORTSTATUS",
            "TOTVAL",
            "WCOHABIT",
            "WEIGHT",
            "WFOSTER22",
            "WICVAL",
            "WKXPNS",
            "WNEWHEAD",
            "WNEWPARENT",
            "WUI_LT15",
            "ID",
        ]
        return (
            person[["SPM_" + column for column in SPM_UNIT_COLUMNS]]
            .groupby(person.SPM_ID)
            .first()
        )


RawCPS = RawCPS()



from openfisca_us.model_api import *


class ut_taxable_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "UT taxable income"
    unit = USD
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        salt_refund = tax_unit("salt_refund_last_year", period)
        ut_subtractions = tax_unit("ut_subtractions_from_income", period)
        ut_total_income = tax_unit("ut_total_income", period)

        return ut_total_income - salt_refund - ut_subtractions



from openfisca_us.model_api import *


class rptc(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Refundable Payroll Tax Credit"
    documentation = "Refundable Payroll Tax Credit for filing unit"
    unit = USD


refundable_payroll_tax_credit = variable_alias(
    "refundable_payroll_tax_credit", rptc
)


class rptc_p(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    documentation = "Refundable Payroll Tax Credit for taxpayer"
    unit = USD


class rptc_s(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    documentation = "Refundable Payroll Tax Credit for spouse"
    unit = USD



from openfisca_us.model_api import *


class non_refundable_american_opportunity_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Non-refundable American Opportunity Credit"
    unit = USD
    documentation = "Value of the non-refundable portion of the American Opportunity Credit"
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/25A#i"

    def formula(tax_unit, period):
        total = tax_unit("american_opportunity_credit", period)
        refundable = tax_unit("refundable_american_opportunity_credit", period)
        return total - refundable


c87668 = variable_alias("c87668", non_refundable_american_opportunity_credit)



from openfisca_us.model_api import *


class county_zip_code(Variable):
    value_type = str
    entity = Household
    label = "County code from ZIP code"
    definition_period = YEAR
    reference = "https://en.wikipedia.org/wiki/ZIP_Code"

    def formula(household, period, parameters):
        # TODO: investigate how consistent this is across the US
        zip_code = household("zip_code", period)
        is_unknown = zip_code == "UNKNOWN"
        zip_code = where(is_unknown, 0, zip_code).astype(int)
        return where(is_unknown, -1, ((zip_code % 1e5) // 1e2).astype(str))



from openfisca_us.model_api import *


class cdcc_rate(Variable):
    value_type = float
    entity = TaxUnit
    label = "CDCC credit rate"
    unit = USD
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/21#a_2"

    def formula(tax_unit, period, parameters):
        cdcc = parameters(period).gov.irs.credits.cdcc
        agi = tax_unit("adjusted_gross_income", period)

        # First phase-out
        phase_out_agi = max_(0, agi - cdcc.phase_out.start)
        percentage_reduction = np.ceil(cdcc.phase_out.rate * phase_out_agi)
        phased_out_rate = max_(
            cdcc.phase_out.min, cdcc.phase_out.max - percentage_reduction
        )

        # Second phase-out
        second_phase_out_agi = max_(0, agi - cdcc.phase_out.second_start)
        second_percentage_reduction = np.ceil(
            cdcc.phase_out.rate * second_phase_out_agi
        )
        return max_(0, phased_out_rate - second_percentage_reduction)



from openfisca_us.model_api import *


class ccdf_income(Variable):
    value_type = float
    entity = SPMUnit
    label = "Income"
    definition_period = YEAR
    unit = USD

    def formula(spm_unit, period, parameters):
        return spm_unit.sum(spm_unit.members("market_income", period))



from openfisca_us.model_api import *


class income_tax_refundable_credits(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Refundable tax credits"
    documentation = "Total refundable income tax credits"
    unit = USD

    def formula(tax_unit, period, parameters):
        credits = parameters(period).gov.irs.credits.refundable
        CONTRIB_CREDITS = ["ecpa_filer_credit", "ecpa_adult_dependent_credit"]
        return add(tax_unit, period, credits + CONTRIB_CREDITS)



from openfisca_us.model_api import *

# Counties in NY state only for now.


class County(Enum):
    ALBANY_NY = "Albany, NY"
    ALLEGANY_NY = "Allegany, NY"
    BRONX_NY = "Bronx, NY"
    BROOME_NY = "Broome, NY"
    CATTARAUGUS_NY = "Cattaraugus, NY"
    CAYUGA_NY = "Cayuga, NY"
    CHAUTAUQUA_NY = "Chautauqua, NY"
    CHEMUNG_NY = "Chemung, NY"
    CHENANGO_NY = "Chenango, NY"
    CLINTON_NY = "Clinton, NY"
    COLUMBIA_NY = "Columbia, NY"
    CORTLAND_NY = "Cortland, NY"
    DELAWARE_NY = "Delaware, NY"
    DUTCHESS_NY = "Dutchess, NY"
    ERIE_NY = "Erie, NY"
    ESSEX_NY = "Essex, NY"
    FRANKLIN_NY = "Franklin, NY"
    FULTON_NY = "Fulton, NY"
    GENESEE_NY = "Genesee, NY"
    GREENE_NY = "Greene, NY"
    HAMILTON_NY = "Hamilton, NY"
    HERKIMER_NY = "Herkimer, NY"
    JEFFERSON_NY = "Jefferson, NY"
    KINGS_NY = "Kings, NY"
    LEWIS_NY = "Lewis, NY"
    LIVINGSTON_NY = "Livingston, NY"
    MADISON_NY = "Madison, NY"
    MONROE_NY = "Monroe, NY"
    MONTGOMERY_NY = "Montgomery, NY"
    NASSAU_NY = "Nassau, NY"
    NEW_YORK_NY = "New York, NY"
    NIAGARA_NY = "Niagara, NY"
    ONEIDA_NY = "Oneida, NY"
    ONONDAGA_NY = "Onondaga, NY"
    ONTARIO_NY = "Ontario, NY"
    ORANGE_NY = "Orange, NY"
    ORLEANS_NY = "Orleans, NY"
    OSWEGO_NY = "Oswego, NY"
    OTSEGO_NY = "Otsego, NY"
    PUTNAM_NY = "Putnam, NY"
    QUEENS_NY = "Queens, NY"
    RENSSELAER_NY = "Rensselaer, NY"
    RICHMOND_NY = "Richmond, NY"
    ROCKLAND_NY = "Rockland, NY"
    ST_LAWRENCE_NY = "St. Lawrence, NY"
    SARATOGA_NY = "Saratoga, NY"
    SCHENECTADY_NY = "Schenectady, NY"
    SCHOHARIE_NY = "Schoharie, NY"
    SCHUYLER_NY = "Schuyler, NY"
    SENECA_NY = "Seneca, NY"
    STEUBEN_NY = "Steuben, NY"
    SUFFOLK_NY = "Suffolk, NY"
    SULLIVAN_NY = "Sullivan, NY"
    TIOGA_NY = "Tioga, NY"
    TOMPKINS_NY = "Tompkins, NY"
    ULSTER_NY = "Ulster, NY"
    WARREN_NY = "Warren, NY"
    WASHINGTON_NY = "Washington, NY"
    WAYNE_NY = "Wayne, NY"
    WESTCHESTER_NY = "Westchester, NY"
    WYOMING_NY = "Wyoming, NY"
    YATES_NY = "Yates, NY"


class county(Variable):
    value_type = Enum
    possible_values = County
    default_value = County.NEW_YORK_NY
    entity = Household
    label = "County"
    definition_period = ETERNITY



from openfisca_us.model_api import *


class snap_earned_income(Variable):
    value_type = float
    entity = SPMUnit
    definition_period = YEAR
    label = "SNAP earned income"
    documentation = (
        "Earned income for calculating the SNAP earned income deduction"
    )
    reference = "https://www.law.cornell.edu/cfr/text/7/273.9#b_1"
    unit = USD

    def formula(spm_unit, period, parameters):
        sources = parameters(period).gov.usda.snap.income.sources.earned
        return aggr(spm_unit, period, sources)



from openfisca_us.model_api import *


class ma_income_tax(Variable):
    value_type = float
    entity = TaxUnit
    label = "MA income tax"
    unit = USD
    documentation = "Massachusetts State income tax."
    definition_period = YEAR
    reference = "https://www.mass.gov/doc/2021-form-1-massachusetts-resident-income-tax-return/download"

    def formula(tax_unit, period, parameters):
        in_ma = tax_unit.household("state_code_str", period) == "MA"
        income_tax_before_credits = tax_unit(
            "ma_income_tax_before_credits", period
        )
        CREDITS = [
            "ma_limited_income_tax_credit",
            "ma_eitc",
            "ma_dependent_credit",
        ]
        credit_value = add(tax_unit, period, CREDITS)
        return in_ma * (income_tax_before_credits - credit_value)



from openfisca_us.model_api import *


class taxbc(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Regular tax before credits"
    documentation = "Regular tax on regular taxable income before credits"
    unit = USD

    def formula(tax_unit, period, parameters):
        capital_gains = parameters(period).gov.irs.capital_gains.brackets
        filing_status = tax_unit("filing_status", period)
        dwks1 = tax_unit("taxable_income", period)

        dwks16 = min_(capital_gains.thresholds["1"][filing_status], dwks1)
        dwks17 = min_(tax_unit("dwks14", period), dwks16)

        dwks20 = dwks16 - dwks17
        lowest_rate_tax = capital_gains.rates["1"] * dwks20
        # Break in worksheet lines
        dwks13 = tax_unit("dwks13", period)
        dwks21 = min_(dwks1, dwks13)
        dwks22 = dwks20
        dwks23 = max_(0, dwks21 - dwks22)
        dwks25 = min_(capital_gains.thresholds["2"][filing_status], dwks1)
        dwks19 = tax_unit("dwks19", period)
        dwks26 = min_(dwks19, dwks20)
        dwks27 = max_(0, dwks25 - dwks26)
        dwks28 = min_(dwks23, dwks27)
        dwks29 = capital_gains.rates["2"] * dwks28
        dwks30 = dwks22 + dwks28
        dwks31 = dwks21 - dwks30
        dwks32 = capital_gains.rates["3"] * dwks31
        # Break in worksheet lines
        dwks33 = min_(
            tax_unit("dwks9", period),
            add(tax_unit, period, ["unrecaptured_section_1250_gain"]),
        )
        dwks10 = tax_unit("dwks10", period)
        dwks34 = dwks10 + dwks19
        dwks36 = max_(0, dwks34 - dwks1)
        dwks37 = max_(0, dwks33 - dwks36)
        dwks38 = 0.25 * dwks37
        # Break in worksheet lines
        dwks39 = dwks19 + dwks20 + dwks28 + dwks31 + dwks37
        dwks40 = dwks1 - dwks39
        dwks41 = 0.28 * dwks40

        # SchXYZ call in Tax-Calculator

        # Separate non-negative taxable income into two non-negative components,
        # doing this in a way so that the components add up to taxable income
        # define pass-through income eligible for PT schedule
        individual_income = parameters(period).gov.irs.income
        e26270 = add(tax_unit, period, ["partnership_s_corp_income"])
        e00900 = add(tax_unit, period, ["self_employment_income"])

        # Determine pass-through and non-pass-through income
        pt_active_gross = e00900 + e26270
        pt_active = pt_active_gross
        pt_active = min_(pt_active, e00900 + e26270)
        pt_taxinc = max_(0, pt_active)
        taxable_income = dwks19

        pt_taxinc = min_(pt_taxinc, taxable_income)
        reg_taxinc = max_(0, taxable_income - pt_taxinc)
        pt_tbase = reg_taxinc

        filing_status = tax_unit("filing_status", period)

        # Initialise regular and pass-through income tax to zero
        reg_tax = 0
        pt_tax = 0
        last_reg_threshold = 0
        last_pt_threshold = 0
        for i in range(1, 7):
            # Calculate rate applied to regular income up to the current
            # threshold (on income above the last threshold)
            reg_threshold = individual_income.bracket.thresholds[str(i)][
                filing_status
            ]
            amount_in_bracket = amount_between(
                reg_taxinc, last_reg_threshold, reg_threshold
            )
            reg_tax += (
                individual_income.bracket.rates[str(i)] * amount_in_bracket
            )
            last_reg_threshold = reg_threshold

            # Calculate rate applied to pass-through income on in the same
            # way, but as treated as if stacked on top of regular income
            # (which is not taxed again)
            pt_threshold = max_(
                individual_income.pass_through.bracket.thresholds[str(i)][
                    filing_status
                ]
                - pt_tbase,
                0,
            )
            pt_tax += individual_income.pass_through.bracket.rates[
                str(i)
            ] * amount_between(pt_taxinc, last_pt_threshold, pt_threshold)
            last_pt_threshold = pt_threshold

        # Calculate regular and pass-through tax above the last threshold
        reg_tax += individual_income.bracket.rates["7"] * max_(
            reg_taxinc - last_reg_threshold, 0
        )
        pt_tax += individual_income.pass_through.bracket.rates["7"] * max_(
            pt_taxinc - last_pt_threshold, 0
        )

        dwks42 = reg_tax + pt_tax

        dwks43 = sum(
            [
                dwks29,
                dwks32,
                dwks38,
                dwks41,
                dwks42,
                lowest_rate_tax,
            ]
        )
        c05200 = tax_unit("income_tax_main_rates", period)
        dwks44 = c05200
        dwks45 = min_(dwks43, dwks44)

        hasqdivltcg = tax_unit("hasqdivltcg", period)

        return where(hasqdivltcg, dwks45, c05200)


regular_tax_before_credits = variable_alias(
    "regular_tax_before_credits", taxbc
)



from openfisca_us.model_api import *


class ma_part_a_taxable_interest_dividend_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "MA Part A taxable income from interest or dividends"
    unit = USD
    definition_period = YEAR
    reference = "https://www.mass.gov/info-details/mass-general-laws-c62-ss-3"

    def formula(tax_unit, period, parameters):
        part_a_taxable_income = tax_unit("ma_part_a_taxable_income", period)
        interest_and_dividends = add(
            tax_unit, period, ["interest_income", "dividend_income"]
        )
        return min_(part_a_taxable_income, interest_and_dividends)



from openfisca_us.model_api import *


class c05200(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Income tax main rates"
    reference = "https://www.law.cornell.edu/uscode/text/26/1"
    unit = USD

    def formula(tax_unit, period, parameters):
        # Separate non-negative taxable income into two non-negative components,
        # doing this in a way so that the components add up to taxable income
        # define pass-through income eligible for PT schedule
        individual_income = parameters(period).gov.irs.income
        e26270 = add(tax_unit, period, ["partnership_s_corp_income"])
        e00900 = add(tax_unit, period, ["self_employment_income"])

        # Determine pass-through and non-pass-through income
        pt_active_gross = e00900 + e26270
        pt_active = pt_active_gross
        pt_active = min_(pt_active, e00900 + e26270)
        pt_taxinc = max_(0, pt_active)
        taxable_income = tax_unit("taxable_income", period)

        pt_taxinc = min_(pt_taxinc, taxable_income)
        reg_taxinc = max_(0, taxable_income - pt_taxinc)
        pt_tbase = reg_taxinc

        filing_status = tax_unit("filing_status", period)

        # Initialise regular and pass-through income tax to zero
        reg_tax = 0
        pt_tax = 0
        last_reg_threshold = 0
        last_pt_threshold = 0
        for i in range(1, 7):
            # Calculate rate applied to regular income up to the current
            # threshold (on income above the last threshold)
            reg_threshold = individual_income.bracket.thresholds[str(i)][
                filing_status
            ]
            reg_tax += individual_income.bracket.rates[
                str(i)
            ] * amount_between(reg_taxinc, last_reg_threshold, reg_threshold)
            last_reg_threshold = reg_threshold

            # Calculate rate applied to pass-through income on in the same
            # way, but as treated as if stacked on top of regular income
            # (which is not taxed again)
            pt_threshold = max_(
                0,
                individual_income.pass_through.bracket.thresholds[str(i)][
                    filing_status
                ]
                - pt_tbase,
            )
            pt_tax += individual_income.pass_through.bracket.rates[
                str(i)
            ] * amount_between(pt_taxinc, last_pt_threshold, pt_threshold)
            last_pt_threshold = pt_threshold

        # Calculate regular and pass-through tax above the last threshold
        reg_tax += individual_income.bracket.rates["7"] * max_(
            reg_taxinc - last_reg_threshold, 0
        )
        pt_tax += individual_income.pass_through.bracket.rates["7"] * max_(
            pt_taxinc - last_pt_threshold, 0
        )
        return reg_tax + pt_tax


income_tax_main_rates = variable_alias("income_tax_main_rates", c05200)



from functools import reduce
from openfisca_core.model_api import *
from openfisca_us.entities import *
from openfisca_us.tools.general import *
from pathlib import Path
from openfisca_us.typing import *
import warnings
from openfisca_us.tools.cloning import get_stored_variables

warnings.filterwarnings("ignore")

REPO = Path(__file__).parent


def all_of_variables(variables: List[str]) -> Formula:
    def formula(entity, period, parameters):
        value = True
        for variable in variables:
            value = value & (add(entity, period, [variable]) > 0)
        return value

    return formula



from openfisca_us.model_api import *


class e07260(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    documentation = "Residential energy credit from Form 5695"
    unit = USD


residential_energy_credit = variable_alias("residential_energy_credit", e07260)



from openfisca_us.model_api import *


class medicaid_rating_area(Variable):
    value_type = int
    entity = Household
    label = "Medicaid rating area"
    definition_period = YEAR

    def formula(household, period, parameters):
        mra = parameters(
            period
        ).gov.hhs.medicaid.geography.medicaid_rating_area
        state = household("state_code_str", period)
        has_defined_mra = pd.Series(state).isin(mra._children)
        state = where(
            has_defined_mra, state, list(mra._children.keys())[0]
        )  # Fill in with any valid State to avoid errors
        three_digit_zip_code = household("three_digit_zip_code", period)
        return where(has_defined_mra, mra[state][three_digit_zip_code], -1)



from openfisca_core.model_api import *
from openfisca_us.entities import *
from openfisca_tools.model_api import *
import numpy as np

USD = "currency-USD"


def tax_unit_non_dep_sum(var, tax_unit, period):
    return tax_unit.sum(
        tax_unit.members(var, period)
        * not_(tax_unit.members("is_tax_unit_dependent", period))
    )


def sum_contained_tax_units(var, population, period):
    tax_unit = population.members.tax_unit.reference_entity
    values = tax_unit(var, period)
    is_head = population.members("is_tax_unit_head", period)
    person_level_values = tax_unit.project(values) * is_head
    return population.sum(person_level_values)


infinity = np.inf
select = np.select
where = np.where

PERCENT = "/1"


def variable_alias(name: str, variable_cls: type) -> type:
    """
    Copy a variable class and return a new class.
    """
    class_dict = dict(variable_cls.__dict__)
    class_dict["formula"] = lambda entity, period: entity(
        variable_cls.__name__, period
    )
    return type(
        name,
        variable_cls.__bases__,
        class_dict,
    )


def taxcalc_read_only_variable(name: str, variable_cls: type) -> type:
    """
    Copy a variable class and return a new class for a tax-calc variable.
    """
    class_dict = dict(variable_cls.__dict__)
    class_dict["formula"] = lambda entity, period: entity(
        variable_cls.__name__, period
    )
    return type(
        name,
        (Variable,),
        dict(
            value_type=variable_cls.value_type,
            entity=variable_cls.entity,
            label=variable_cls.label + " (Tax-Calculator)",
            definition_period=variable_cls.definition_period,
            unit=variable_cls.unit,
            documentation=variable_cls.documentation
            + " This is a read-only copy variable, matching the corresponding variable in the open-source US federal tax model Tax-Calculator.",
        ),
    )


def sum_among_non_dependents(variable: str) -> Callable:
    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum(variable, tax_unit, period)

    return formula


def spouse(person: Population, period: int, variable: str) -> ArrayLike:
    values = person(variable, period)
    return (person.marital_unit.sum(values) - values).astype(values.dtype)


def in_state(state):
    def is_eligible(population, period, parameters):
        return population.household("state_code_str", period) == state

    return is_eligible



from openfisca_us.model_api import *


class short_term_capital_gains(Variable):
    value_type = float
    entity = Person
    label = "Short-term capital gains"
    unit = USD
    definition_period = YEAR



from openfisca_us.model_api import *


class ira_income(Variable):
    value_type = float
    entity = Person
    label = "IRA income"
    unit = USD
    documentation = "Income from all Individual Retirement Accounts."
    definition_period = YEAR

    formula = sum_of_variables(["tax_exempt_ira_income", "taxable_ira_income"])



import logging
from openfisca_tools.data import PublicDataset
import h5py
from openfisca_us.data.datasets.cps.raw_cps import RawCPS
from openfisca_us.data.storage import OPENFISCA_US_MICRODATA_FOLDER
from pandas import DataFrame, Series
import numpy as np
import pandas as pd


class CPS(PublicDataset):
    name = "cps"
    label = "CPS"
    model = "openfisca_us"
    folder_path = OPENFISCA_US_MICRODATA_FOLDER

    url_by_year = {
        2020: "https://github.com/PolicyEngine/openfisca-us/releases/download/cps-v0/cps_2020.h5"
    }

    def generate(self, year: int):
        """Generates the Current Population Survey dataset for OpenFisca-US microsimulations.
        Technical documentation and codebook here: https://www2.census.gov/programs-surveys/cps/techdocs/cpsmar21.pdf

        Args:
            year (int): The year of the Raw CPS to use.
        """

        # Prepare raw CPS tables
        year = int(year)
        if year not in RawCPS.years:
            logging.info(f"Generating raw CPS for year {year}.")
            RawCPS.generate(year)

        raw_data = RawCPS.load(year)
        cps = h5py.File(self.file(year), mode="w")

        person, tax_unit, family, spm_unit, household = [
            raw_data[entity]
            for entity in (
                "person",
                "tax_unit",
                "family",
                "spm_unit",
                "household",
            )
        ]

        add_id_variables(cps, person, tax_unit, family, spm_unit, household)
        add_personal_variables(cps, person)
        add_personal_income_variables(cps, person)
        add_spm_variables(cps, spm_unit)
        add_household_variables(cps, household)

        raw_data.close()
        cps.close()


def add_id_variables(
    cps: h5py.File,
    person: DataFrame,
    tax_unit: DataFrame,
    family: DataFrame,
    spm_unit: DataFrame,
    household: DataFrame,
) -> None:
    """Add basic ID and weight variables.

    Args:
        cps (h5py.File): The CPS dataset file.
        person (DataFrame): The person table of the ASEC.
        tax_unit (DataFrame): The tax unit table created from the person table
            of the ASEC.
        family (DataFrame): The family table of the ASEC.
        spm_unit (DataFrame): The SPM unit table created from the person table
            of the ASEC.
        household (DataFrame): The household table of the ASEC.
    """
    # Add primary and foreign keys
    cps["person_id"] = person.PH_SEQ * 100 + person.P_SEQ
    cps["family_id"] = family.FH_SEQ * 10 + family.FFPOS
    cps["household_id"] = household.H_SEQ
    cps["person_tax_unit_id"] = person.TAX_ID
    cps["person_spm_unit_id"] = person.SPM_ID
    cps["tax_unit_id"] = tax_unit.TAX_ID
    cps["spm_unit_id"] = spm_unit.SPM_ID
    cps["person_household_id"] = person.PH_SEQ
    cps["person_family_id"] = person.PH_SEQ * 10 + person.PF_SEQ

    # Add weights
    # Weights are multiplied by 100 to avoid decimals
    cps["person_weight"] = person.A_FNLWGT / 1e2
    cps["family_weight"] = family.FSUP_WGT / 1e2

    # Tax unit weight is the weight of the containing family.
    family_weight = Series(
        cps["family_weight"][...], index=cps["family_id"][...]
    )
    person_family_id = cps["person_family_id"][...]
    persons_family_weight = Series(family_weight[person_family_id])
    cps["tax_unit_weight"] = persons_family_weight.groupby(
        cps["person_tax_unit_id"][...]
    ).first()

    cps["spm_unit_weight"] = spm_unit.SPM_WEIGHT / 1e2

    cps["household_weight"] = household.HSUP_WGT / 1e2

    # Marital units

    marital_unit_id = person.PH_SEQ * 1e6 + np.maximum(
        person.A_LINENO, person.A_SPOUSE
    )

    # marital_unit_id is not the household ID, zero padded and followed
    # by the index within household (of each person, or their spouse if
    # one exists earlier in the survey).

    marital_unit_id = Series(marital_unit_id).rank(
        method="dense"
    )  # Simplify to a natural number sequence with repetitions [0, 1, 1, 2, 3, ...]

    cps["person_marital_unit_id"] = marital_unit_id.values
    cps["marital_unit_id"] = marital_unit_id.drop_duplicates().values


def add_personal_variables(cps: h5py.File, person: DataFrame) -> None:
    """Add personal demographic variables.

    Args:
        cps (h5py.File): The CPS dataset file.
        person (DataFrame): The CPS person table.
    """

    # The CPS edits age as follows:
    # 0-79 => 0-79
    # 80-84  => 80
    # 85+ => 85
    # We assign the 80 ages randomly between 80 and 85
    # to avoid unrealistically bunching at 80.
    cps["age"] = np.where(
        person.A_AGE.between(80, 85),
        80 + 5 * np.random.rand(len(person)),
        person.A_AGE,
    )
    # A_SEX is 1 -> male, 2 -> female.
    cps["is_female"] = person.A_SEX == 2

    def children_per_parent(col: str) -> pd.DataFrame:
        """Calculate number of children in the household using parental
            pointers.

        Args:
            col (str): Either PEPAR1 and PEPAR2, which correspond to A_LINENO
            of the person's first and second parent in the household,
            respectively.
        """
        return (
            person[person[col] > 0]
            .groupby(["PH_SEQ", col])
            .size()
            .reset_index()
            .rename(columns={col: "A_LINENO", 0: "children"})
        )

    # Aggregate to parent.
    res = (
        pd.concat(
            [children_per_parent("PEPAR1"), children_per_parent("PEPAR2")]
        )
        .groupby(["PH_SEQ", "A_LINENO"])
        .children.sum()
        .reset_index()
    )
    tmp = person[["PH_SEQ", "A_LINENO"]].merge(
        res, on=["PH_SEQ", "A_LINENO"], how="left"
    )
    cps["own_children_in_household"] = tmp.children.fillna(0)

    cps["has_marketplace_health_coverage"] = person.MRK == 1


def add_personal_income_variables(cps: h5py.File, person: DataFrame):
    """Add income variables.

    Args:
        cps (h5py.File): The CPS dataset file.
        person (DataFrame): The CPS person table.
    """
    cps["employment_income"] = person.WSAL_VAL
    cps["interest_income"] = person.INT_VAL
    cps["self_employment_income"] = person.SEMP_VAL
    cps["farm_income"] = person.FRSE_VAL
    cps["dividend_income"] = person.DIV_VAL
    cps["rental_income"] = person.RNT_VAL
    cps["social_security"] = person.SS_VAL
    cps["unemployment_compensation"] = person.UC_VAL
    cps["pension_income"] = person.PNSN_VAL + person.ANN_VAL
    cps["alimony_income"] = (person.OI_OFF == 20) * person.OI_VAL
    cps["tanf_reported"] = person.PAW_VAL
    cps["ssi_reported"] = person.SSI_VAL
    cps["pension_contributions"] = person.RETCB_VAL
    cps[
        "long_term_capital_gains"
    ] = person.CAP_VAL  # Assume all CPS capital gains are long-term
    cps["receives_wic"] = person.WICYN == 1


def add_spm_variables(cps: h5py.File, spm_unit: DataFrame) -> None:
    SPM_RENAMES = dict(
        spm_unit_total_income_reported="SPM_TOTVAL",
        snap_reported="SPM_SNAPSUB",
        spm_unit_capped_housing_subsidy_reported="SPM_CAPHOUSESUB",
        free_school_meals_reported="SPM_SCHLUNCH",
        spm_unit_energy_subsidy_reported="SPM_ENGVAL",
        spm_unit_wic_reported="SPM_WICVAL",
        spm_unit_payroll_tax_reported="SPM_FICA",
        spm_unit_federal_tax_reported="SPM_FEDTAX",
        spm_unit_state_tax_reported="SPM_STTAX",
        spm_unit_work_childcare_expenses="SPM_CAPWKCCXPNS",
        spm_unit_medical_expenses="SPM_MEDXPNS",
        spm_unit_spm_threshold="SPM_POVTHRESHOLD",
        spm_unit_net_income_reported="SPM_RESOURCES",
        childcare_expenses="SPM_CHILDCAREXPNS",
    )

    for openfisca_variable, asec_variable in SPM_RENAMES.items():
        cps[openfisca_variable] = spm_unit[asec_variable]

    cps["reduced_price_school_meals_reported"] = (
        cps["free_school_meals_reported"][...] * 0
    )


def add_household_variables(cps: h5py.File, household: DataFrame) -> None:
    cps["fips"] = household.GESTFIPS


CPS = CPS()



from openfisca_us.model_api import *


class basic_income(Variable):
    value_type = float
    entity = Person
    label = "Basic income"
    unit = USD
    documentation = "Total basic income payments for this person. Phase-outs as an equal percentage to all tax unit members."
    definition_period = YEAR

    def formula(person, period, parameters):
        basic_income = person("bi_before_phase_out", period)
        tax_unit = person.tax_unit
        tax_unit_basic_income = tax_unit.sum(basic_income)
        tax_unit_phase_out = tax_unit("bi_phase_out", period)
        percent_reduction = where(
            tax_unit_basic_income > 0,
            tax_unit_phase_out / tax_unit_basic_income,
            0,
        )
        return basic_income * (1 - percent_reduction)



from openfisca_us.model_api import *


class cdcc_refund(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Child/dependent care refundable credit"
    unit = USD
    documentation = "Refundable credit for child and dependent care expenses from Form 2441"
    reference = "https://www.law.cornell.edu/uscode/text/26/21#g_1"

    def formula(tax_unit, period, parameters):
        cdcc = parameters(period).gov.irs.credits.cdcc
        if cdcc.refundable:
            return tax_unit("cdcc", period)
        else:
            return 0



from openfisca_us.model_api import *


class state_income_tax(Variable):
    value_type = float
    entity = TaxUnit
    label = "State income tax"
    unit = USD
    definition_period = YEAR

    def formula_2021(tax_unit, period, parameters):
        return tax_unit("ma_income_tax", period)



from openfisca_us.model_api import *
from openfisca_us.variables.gov.irs.credits.ctc.maximum.individual.ctc_child_individual_maximum import (
    ctc_child_individual_maximum,
)
from openfisca_us.variables.gov.irs.credits.ctc.maximum.individual.ctc_adult_individual_maximum import (
    ctc_adult_individual_maximum,
)
from openfisca_us.parameters import default_parameters


class ctc_reduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "CTC reduction from income"
    unit = USD
    documentation = "Reduction of the total CTC due to income."
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        income = tax_unit("adjusted_gross_income", period)
        ctc = parameters(period).gov.irs.credits.ctc
        filing_status = tax_unit("filing_status", period)
        income_over_threshold = max_(
            0, income - ctc.phase_out.threshold[filing_status]
        )
        reduction = ctc.phase_out.rate * income_over_threshold
        maximum_ctc = tax_unit("ctc_maximum", period)
        return min_(reduction, maximum_ctc)

    # TCJA's phase-out changes are purely parametric so don't require structural reform.

    def formula_2021(tax_unit, period, parameters):
        # The ARPA CTC has two phase-outs: the original, and a new phase-out
        # applying before and only to the increase in the maximum CTC under ARPA.

        income = tax_unit("adjusted_gross_income", period)
        ctc = parameters(period).gov.irs.credits.ctc
        filing_status = tax_unit("filing_status", period)
        income_over_threshold = max_(
            0, income - ctc.phase_out.threshold[filing_status]
        )
        reduction = ctc.phase_out.rate * income_over_threshold
        maximum_ctc = tax_unit("ctc_maximum", period)

        # Calculate the original phase-out
        original_phase_out = min_(reduction, maximum_ctc)

        # Calculate the income used to assess the new phase-out
        income_over_arpa_threshold = max_(
            0, income - ctc.phase_out.arpa.threshold[filing_status]
        )
        arpa_phase_out_max_reduction = (
            ctc.phase_out.arpa.rate * income_over_arpa_threshold
        )

        # Calculate the increase - do this by finding the original CTC if
        # ARPA had not applied - this year's variables, last year's parameters.
        no_arpa_parameters = default_parameters.clone()
        old_ctc = parameters(period.last_year).gov.irs.credits.ctc
        no_arpa_ctc = no_arpa_parameters.gov.irs.credits.ctc
        no_arpa_ctc.child.young.increase.update(value=0, period=period)
        no_arpa_ctc.child.amount.update(
            value=old_ctc.child.amount, period=period
        )

        ctc_without_arpa = tax_unit.sum(
            ctc_child_individual_maximum.formula(
                tax_unit.members, period, no_arpa_parameters
            )
            + ctc_adult_individual_maximum.formula_2018(
                tax_unit.members, period, no_arpa_parameters
            )
        )

        arpa_increase = maximum_ctc - ctc_without_arpa

        arpa_phase_out_range = (
            ctc.phase_out.threshold[filing_status]
            - ctc.phase_out.arpa.threshold[filing_status]
        )

        # Apply the phase-out
        arpa_reduction_max = min_(
            arpa_increase,
            ctc.phase_out.arpa.rate * arpa_phase_out_range,
        )

        arpa_reduction = min_(arpa_phase_out_max_reduction, arpa_reduction_max)

        return original_phase_out + arpa_reduction

    formula_2022 = formula



from openfisca_us.data import ACS
import pytest
from openfisca_us import Microsimulation

ACS_YEARS = []


@pytest.mark.dependency(name="acs")
@pytest.mark.parametrize("year", ACS_YEARS)
def test_acs_dataset_generates(year):
    ACS.generate(year)


@pytest.mark.dependency(depends=["acs"])
@pytest.mark.parametrize("year", ACS_YEARS)
def test_acs_openfisca_us_compatible(year):
    Microsimulation(dataset=ACS, year=year).calc("employment_income")



from openfisca_us.model_api import *


class qualified_business_income_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Qualified business income deduction"
    unit = USD
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/199A#b_1"

    def formula(tax_unit, period, parameters):
        max_qbid = tax_unit("maximum_qbid", period)
        limit = tax_unit("qbid_limit", period)
        return min_(max_qbid, limit)



from openfisca_us.model_api import *


class taxable_ira_income(Variable):
    value_type = float
    entity = Person
    label = "Taxable IRA income."
    unit = USD
    documentation = "Income from traditional IRAs."
    definition_period = YEAR



from openfisca_us.model_api import *


class loss_ald(Variable):
    value_type = float
    entity = TaxUnit
    label = "Business loss ALD"
    unit = USD
    documentation = (
        "Above-the-line deduction from gross income for business losses."
    )
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/165"

    def formula(tax_unit, period, parameters):
        filing_status = tax_unit("filing_status", period)
        max_loss = parameters(period).gov.irs.ald.loss.max[filing_status]
        personal_self_employment_loss = max_(
            -tax_unit.members("self_employment_income", period), 0
        )
        tax_unit_self_employment_losses = tax_unit.sum(
            personal_self_employment_loss
        )
        capital_loss = max_(tax_unit("maximum_capital_loss", period), 0)
        return min_(max_loss, tax_unit_self_employment_losses + capital_loss)



from openfisca_us.model_api import *


class bi_phase_out(Variable):
    value_type = float
    entity = TaxUnit
    label = "Basic income phase-out"
    definition_period = YEAR
    unit = USD

    def formula(tax_unit, period, parameters):
        basic_income = add(tax_unit, period, ["bi_before_phase_out"])
        phase_out = parameters(
            period
        ).contrib.ubi_center.basic_income.phase_out
        agi = tax_unit("adjusted_gross_income", period)
        mars = tax_unit("mars", period)
        threshold = phase_out.threshold[mars]
        income_over_threshold = max_(0, agi - threshold)
        return min_(income_over_threshold * phase_out.rate, basic_income)



from openfisca_us.model_api import *


class tax_exempt_ira_income(Variable):
    value_type = float
    entity = Person
    label = "Tax-exempt IRA income."
    unit = USD
    documentation = "Income from Roth IRAs."
    definition_period = YEAR



from openfisca_us.model_api import *


class c07100(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Income tax non-refundable credits"
    documentation = (
        "Total non-refundable credits used to reduce positive tax liability"
    )
    unit = USD

    def formula(tax_unit, period, parameters):
        credits = parameters(period).gov.irs.credits.non_refundable
        return add(tax_unit, period, credits)


income_tax_non_refundable_credits = variable_alias(
    "income_tax_non_refundable_credits", c07100
)


class income_tax_capped_non_refundable_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "Non-refundable tax credits"
    unit = USD
    documentation = "Capped value of non-refundable tax credits"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        return min_(
            tax_unit("income_tax_before_credits", period),
            tax_unit("income_tax_non_refundable_credits", period),
        )



from openfisca_us.model_api import *


class ctc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Child Tax Credit"
    unit = USD
    documentation = "Total value of the non-refundable and refundable portions of the Child Tax Credit."
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/24#a"

    def formula(tax_unit, period, parameters):
        maximum_amount = tax_unit("ctc_maximum", period)
        reduction = tax_unit("ctc_reduction", period)
        return maximum_amount - reduction



from openfisca_us.model_api import *


class three_digit_zip_code(Variable):
    value_type = str
    entity = Household
    label = "Three-digit zipcode"
    definition_period = YEAR

    def formula(household, period, parameters):
        zip_code = household("zip_code", period)
        if any(zip_code == "UNKNOWN"):
            # We don't have full zip codes - randomly assign zip codes
            population_data = parameters(
                period
            ).demographic.geography.population_by_three_digit_zip_code
            zip_codes = np.array(list(population_data._children.keys()))
            populations = np.array(list(population_data._children.values()))
            np.random.seed(0)
            zip_code = (
                np.random.choice(
                    zip_codes,
                    size=len(zip_code),
                    p=populations / populations.sum(),
                )
                .astype(int)
                .astype(str)
            )
        else:
            zip_code = (zip_code.astype(int) // 1e2).astype(int).astype(str)
        return np.char.zfill(zip_code, 3)



from openfisca_us.model_api import *


class c62100(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "AMT taxable income"
    unit = USD
    documentation = "Alternative Minimum Tax (AMT) taxable income"

    def formula(tax_unit, period, parameters):
        # Form 6251, Part I
        adjusted_gross_income = tax_unit("adjusted_gross_income", period)
        e00700 = add(tax_unit, period, ["salt_refund_income"])
        c62100_if_no_standard = (
            adjusted_gross_income
            - e00700
            - tax_unit("taxable_income_deductions_if_itemizing", period)
            + max_(
                0,
                min_(
                    tax_unit("medical_expense_deduction", period),
                    0.025 * adjusted_gross_income,
                ),
            )
            + tax_unit("salt_deduction", period)
        )
        c62100 = where(
            tax_unit("standard_deduction", period) == 0,
            c62100_if_no_standard,
            adjusted_gross_income - e00700,
        ) + tax_unit(
            "amt_non_agi_income", period
        )  # add income not in AGI but considered income for AMT
        amt = parameters(period).gov.irs.income.amt
        filing_status = tax_unit("filing_status", period)
        separate_addition = max_(
            0,
            min_(
                amt.exemption.amount[filing_status],
                amt.exemption.phase_out.rate
                * (c62100 - amt.exemption.separate_limit),
            ),
        ) * (filing_status == filing_status.possible_values.SEPARATE)
        return c62100 + separate_addition


amt_income = variable_alias("amt_income", c62100)


class c09600(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Alternative Minimum Tax"
    unit = USD
    documentation = "Alternative Minimum Tax (AMT) liability"

    def formula(tax_unit, period, parameters):
        c62100 = tax_unit("c62100", period)
        # Form 6251, Part II top
        amt = parameters(period).gov.irs.income.amt
        phase_out = amt.exemption.phase_out
        filing_status = tax_unit("filing_status", period)
        line29 = max_(
            0,
            (
                amt.exemption.amount[filing_status]
                - phase_out.rate
                * max_(0, c62100 - phase_out.start[filing_status])
            ),
        )
        age_head = tax_unit("age_head", period)
        child = amt.exemption.child
        young_head = (age_head != 0) & (age_head < child.max_age)
        no_or_young_spouse = tax_unit("age_spouse", period) < child.max_age
        line29 = where(
            young_head & no_or_young_spouse,
            min_(line29, tax_unit("filer_earned", period) + child.amount),
            line29,
        )
        line30 = max_(0, c62100 - line29)
        brackets = amt.brackets
        amount_over_threshold = line30 - brackets.thresholds["1"] / tax_unit(
            "sep", period
        )
        line3163 = brackets.rates["1"] * line30 + brackets.rates["2"] * max_(
            0, amount_over_threshold
        )
        dwks10, dwks13, dwks14, dwks19, e24515 = [
            add(tax_unit, period, [variable])
            for variable in [
                "dwks10",
                "dwks13",
                "dwks14",
                "dwks19",
                "unrecaptured_section_1250_gain",
            ]
        ]
        form_6251_part_iii_required = np.any(
            [
                variable > 0
                for variable in [
                    dwks10,
                    dwks13,
                    dwks14,
                    dwks19,
                    e24515,
                ]
            ]
        )

        # Complete Form 6251, Part III

        line37 = dwks13
        line38 = e24515
        line39 = min_(line37 + line38, dwks10)
        line40 = min_(line30, line39)
        line41 = max_(0, line30 - line40)
        amount_over_threshold = max_(
            0, line41 - amt.brackets.thresholds["1"] / tax_unit("sep", period)
        )
        line42 = (
            amt.brackets.rates["1"] * line41
            + amt.brackets.rates["2"] * amount_over_threshold
        )
        line44 = dwks14
        cg = amt.capital_gains.brackets
        line45 = max_(0, cg.thresholds["1"][filing_status] - line44)
        line46 = min_(line30, line37)
        line47 = min_(line45, line46)
        cgtax1 = line47 * cg.rates["1"]
        line48 = line46 - line47
        line51 = dwks19
        line52 = line45 + line51
        line53 = max_(0, cg.thresholds["2"][filing_status] - line52)
        line54 = min_(line48, line53)
        cgtax2 = line54 * cg.rates["2"]
        line56 = line47 + line54
        line57 = where(line41 == line56, 0, line46 - line56)
        linex2 = where(line41 == line56, 0, max_(0, line54 - line48))
        cgtax3 = line57 * cg.rates["3"]
        line61 = where(
            line38 == 0,
            0,
            0.25 * max_(0, (line30 - line41 - line56 - line57 - linex2)),
        )
        line62 = line42 + cgtax1 + cgtax2 + cgtax3 + line61
        line64 = min_(line3163, line62)
        line31 = where(form_6251_part_iii_required, line64, line3163)
        foreign_tax_credit = tax_unit("foreign_tax_credit", period)

        # Form 6251, Part II bottom
        line32 = where(
            tax_unit("amt_form_completed", period),
            tax_unit("foreign_tax_credit", period),
            foreign_tax_credit,
        )
        line33 = line31 - line32
        return max_(
            0,
            line33
            - max_(
                0,
                (
                    tax_unit("taxbc", period)
                    - foreign_tax_credit
                    - tax_unit("c05700", period)
                ),
            ),
        )


alternative_minimum_tax = variable_alias("alternative_minimum_tax", c09600)



from openfisca_us.model_api import *


class FilingStatus(Enum):
    SINGLE = "Single"
    JOINT = "Joint"
    SEPARATE = "Separate"
    HEAD_OF_HOUSEHOLD = "Head of household"
    WIDOW = "Widow(er)"


class filing_status(Variable):
    value_type = Enum
    entity = TaxUnit
    possible_values = FilingStatus
    default_value = FilingStatus.SINGLE
    definition_period = YEAR
    label = "Filing status for the tax unit"

    def formula(tax_unit, period, parameters):
        has_spouse = add(tax_unit, period, ["is_tax_unit_spouse"]) > 0
        has_dependents = tax_unit("tax_unit_dependents", period) > 0
        return select(
            [has_spouse, has_dependents, True],
            [
                FilingStatus.JOINT,
                FilingStatus.HEAD_OF_HOUSEHOLD,
                FilingStatus.SINGLE,
            ],
        )


# For Tax-Calculator.
mars = variable_alias("mars", filing_status)



from openfisca_us.model_api import *


class ma_part_a_taxable_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "MA Part A taxable income"
    unit = USD
    definition_period = YEAR
    reference = "https://www.mass.gov/info-details/mass-general-laws-c62-ss-3"

    def formula(tax_unit, period, parameters):
        part_a_agi = tax_unit("ma_part_a_agi", period)
        part_b_agi = tax_unit("ma_part_b_agi", period)
        part_b_deductions = tax_unit(
            "ma_part_b_taxable_income_deductions", period
        )
        remaining_part_b_deductions = max_(0, part_b_deductions - part_b_agi)
        return max_(0, part_a_agi - remaining_part_b_deductions)



from openfisca_us.model_api import *


class niit(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Net Investment Income Tax"
    reference = "https://www.law.cornell.edu/uscode/text/26/1411"
    unit = USD

    def formula(tax_unit, period, parameters):
        NII_ELEMENTS = [
            "taxable_interest_income",
            "ordinary_dividend_income",
            "c01000",
            "rental_income",
        ]
        nii = max_(0, add(tax_unit, period, NII_ELEMENTS))
        niit = parameters(period).gov.irs.investment.net_inv_inc_tax
        threshold = niit.threshold[tax_unit("filing_status", period)]
        base = min_(
            nii, max_(0, tax_unit("adjusted_gross_income", period) - threshold)
        )
        return niit.rate * base



from openfisca_us.model_api import *


class ut_taxpayer_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "UT taxpayer credit"
    unit = USD
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        ut_taxpayer_credit_max = tax_unit("ut_taxpayer_credit_max", period)
        ut_taxpayer_credit_reduction = tax_unit(
            "ut_taxpayer_credit_reduction", period
        )
        return ut_taxpayer_credit_max - ut_taxpayer_credit_reduction



from openfisca_us.model_api import *


class ctc_limiting_tax_liability(Variable):
    value_type = float
    entity = TaxUnit
    label = "CTC-limiting tax liability"
    unit = USD
    documentation = "The tax liability used to determine the maximum amount of the non-refundable CTC."
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        non_refundable_credits = parameters(
            period
        ).gov.irs.credits.non_refundable
        total_credits = sum(
            [
                tax_unit(credit, period)
                for credit in non_refundable_credits
                if credit != "non_refundable_ctc"
            ]
        )
        return tax_unit("income_tax_before_credits", period) - total_credits



from openfisca_us.model_api import *


class ccdf_county_cluster(Variable):
    value_type = int
    entity = Household
    label = "County cluster for CCDF"
    definition_period = YEAR

    def formula(household, period, parameters):
        county = household("county", period).decode_to_str()
        cluster_mapping = parameters(period).gov.hhs.ccdf.county_cluster
        return cluster_mapping[county]



from openfisca_us.model_api import *


class tax_unit_childcare_expenses(Variable):
    value_type = float
    entity = TaxUnit
    label = "Childcare expenses"
    unit = USD
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        # Distribute the childcare expenses evenly across the SPM unit's members.
        spm_unit = tax_unit.spm_unit
        spm_unit_childcare = spm_unit("childcare_expenses", period)
        spm_unit_count_children = add(spm_unit, period, ["is_child"])
        tax_unit_count_children = add(tax_unit, period, ["is_child"])
        child_ratio = where(
            spm_unit_count_children > 0,
            tax_unit_count_children / spm_unit_count_children,
            0,
        )
        return spm_unit_childcare * child_ratio



from openfisca_us.model_api import *


class interest_income(Variable):
    value_type = float
    entity = Person
    label = "Interest income"
    unit = USD
    definition_period = YEAR

    formula = sum_of_variables(
        ["taxable_interest_income", "tax_exempt_interest_income"]
    )



from openfisca_us.model_api import *


class taxsim_state(Variable):
    value_type = float
    entity = TaxUnit
    label = "State code"
    documentation = 'SOI codes. These run from 1 for Alabama to 51 for Wyoming and are not the Census or PSID codes. See state list,and also item two above.). Use zero for "no state tax calculation"'
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        return 0



from openfisca_us.model_api import *


class c09200(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    unit = USD
    label = "Income tax before refundable credits"
    documentation = "Income tax liability (including othertaxes) after non-refundable credits are used, but before refundable credits are applied"

    def formula(tax_unit, period, parameters):
        income_tax_bc = tax_unit("income_tax_before_credits", period)
        capped_credits = tax_unit(
            "income_tax_capped_non_refundable_credits", period
        )
        return income_tax_bc - capped_credits


income_tax_before_refundable_credits = variable_alias(
    "income_tax_before_refundable_credits", c09200
)



from openfisca_us.model_api import *


class tax_unit_count_dependents(Variable):
    value_type = int
    entity = TaxUnit
    label = "Number of dependents"
    definition_period = YEAR

    formula = sum_of_variables(["is_tax_unit_dependent"])



from openfisca_us.model_api import *


class second_lowest_silver_plan_cost(Variable):
    value_type = float
    entity = TaxUnit
    label = "Second-lowest silver plan cost"
    unit = USD
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        person = tax_unit.members
        household = person.household
        area = household("medicaid_rating_area", period)
        state = household("state_code_str", period)
        slspc = parameters(
            period
        ).gov.hhs.medicaid.geography.second_lowest_silver_plan_cost
        age = person("age", period)
        age_code = select(
            [
                age < 21,
                age < 64,
                age >= 64,
            ],
            [
                "0-20",
                age.astype(int).astype(str),
                "64+",
            ],
        )
        eligible = person.tax_unit("is_ptc_eligible", period)
        per_person_cost = index_(
            into=slspc,
            indices=[state, area, age_code],
            where=eligible,
        )
        return tax_unit.sum(per_person_cost)
